/**
 * Diccionario de correspondencia de colores
 */
const colores = {
    "platform" : {
        Nintendo64:         "#E1A84E",
        PlayStation:        "#4A7DCE",
        PlayStation3:       "#6666e0",
        Dreamcast:          "#D47A45",
        Xbox360:            "#8BBF61",
        Wii:                "#5EB8EC",
        XboxOne:            "#629A62",
        Switch:             "#C84D4D",
        PlayStation2:       "#3A66A1",
        PlayStation4:       "#3D76C9",
        GameCube:           "#8B5DB7",
        Xbox:               "#619A61",
        PC:                 "#9A9A9A",
        WiiU:               "#63B6D5",
        GameBoyAdvance:     "#8574B5",
        "3DS":              "#C1536E",
        DS:                 "#BBBBBB",
        PlayStationVita:    "#5A7EB0",
        PSP:                "#5454f0",
        XboxSeriesX:        "#4F5B4A",
        PlayStation5:       "#C5D1E3",
        Stadia:             "#D45A3B"
    },
    "developer" : {
        Capcom:             "#C75450",
        TelltaleGames:      "#6A9CCB",
        EASports:           "#65A879",
        Nintendo:           "#D84840",
        Konami:             "#C25C74",
        EACanada:           "#518E85",
        UbisoftMontreal:    "#5A78B8",
        Ubisoft:            "#4C96C9",
        SquareEnix:         "#D48E47",
        OmegaForce:         "#A17367",
        ElectronicArts:     "#7A9098",
        EATiburon:          "#68A8B9",
        Codemasters:        "#84A85B",
        Sega:               "#5272A5",
        BandaiNamcoGames:   "#D69444",
        VisualConcepts:     "#6488C0",
        "Traveller'sTales": "#9569A1",
        VicariousVisions:   "#A3A658",
        Namco:              "#C87478"
    }
}


let selection = "platform";
let scatter;

let elegibleData;

let xScale;
let radiusScale;
let yScale;
let xAxis;
let yAxis;
let transform;

let min_x = 0;
let min_y = 0;
let max_x = 10;
let max_y = 10;

const width = 800;
const height = 400;
const margin_x = 62;
const margin_y = 60;
const right_margin = 300;
const rectAparitionTime = 150;
const radiusTransition = 300;
const maxBubbleCount = 50;


const svg = d3.select('.main_view').insert('svg', '.filters-container')
    .attr("width", width + 2 * margin_x).attr("height", height + 2 * margin_y);

svg.append('defs').append('clipPath')
    .attr("id", "content-clip")
    .append('rect')
    .attr('width', width - right_margin)
    .attr('height', height)
    .attr('x', margin_x)
    .attr('y', margin_y);

const contentGroup = svg.append('g')
    .attr('clip-path', 'url(#content-clip)');

const translate_group = svg.append('g').attr("transform", `translate(${margin_x},${margin_y})`);

function createAxis() {
    xAxis = d3.axisBottom(xScale).ticks(10);
    yAxis = d3.axisLeft(yScale).ticks(10);

    translate_group.append("g")
        .attr("class", "title")
        .attr("transform", "translate(10, -20)")
        .append('text')
        .style("font-family", "Georgia, serif")
        .style("font-size", "26px")
        .style("font-weight", "bold")
        .text("Críticas en los Videojuegos: Usuarios y Expertos")


    translate_group.append("g")
        .attr("class", "xAxis")
        .attr("transform", `translate(0,${height})`)
        .call(xAxis);

    translate_group.append('text')
        .attr('class', 'xAxisLabel')
        .attr('x', width / 10)
        .attr('y', height + 40)
        .style('font-weight', 'bold')
        .style('font-size', '20px')
        .style('font-family', 'Georgia, serif')
        .text("Calificación dada por usuarios")

    translate_group.append("g")
        .attr("class", "yAxis")
        .call(yAxis);

    translate_group.append('text')
        .attr('class', 'yAxisLabel')
        .attr('x', -height*4/5)
        .attr('y', -30)
        .style('font-weight', 'bold')
        .style('font-size', '20px')
        .style('font-family', 'Georgia, serif')
        .attr('transform', 'rotate(-90)')
        .text("Calificación dada por expertos")

    contentGroup.selectAll("line.horizontal-grid")
        .data(yScale.ticks(10))
        .enter()
        .append("line")
        .attr("class", "horizontal-grid")
        .attr("x1", 0)
        .attr("x2", width)
        .attr("y1", d => yScale(d) + margin_y)
        .attr("y2", d => yScale(d) + margin_y)
        .attr("stroke", "gray")
        .attr("stroke-width", 0.5);

    contentGroup.selectAll("line.vertical-grid")
        .data(xScale.ticks(10))
        .enter()
        .append("line")
        .attr("class", "vertical-grid")
        .attr("x1", d => xScale(d) + margin_x)
        .attr("x2", d => xScale(d) + margin_x)
        .attr("y1", 0)
        .attr("y2", height + margin_y)
        .attr("stroke", "gray")
        .attr("stroke-width", 0.5);

    translate_group.append('defs').append('clipPath')
        .attr("id", "clip")
        .append('rect')
        .attr('width', width)
        .attr('height', height)
        .attr('x', 0)
        .attr('y', 0);

    scatter = contentGroup.append('g');
}

d3.csv('files/games-data.csv').then((data) => {

    data.forEach(elem => {
        elem["user score"] = +elem["user score"];
        if(!elem["user score"]) elem.critics = -14;
        // Datos que no se puedan representar se eliminan del gráfico así
        elem.score = +elem.score / 10;
        // Los datos estan entre 0 y 100, asi que normalizo a entre 0 y 10
        elem.critics = +elem.critics;
        elem.users = +elem.users;
        elem.genre = [...new Set(elem.genre.split(','))];
        elem.id = elem.name.replaceAll(/[ :'"()\[\]]*/g, '') + "-" + elem.platform;
    });


    xScale = d3.scaleLinear().domain([min_x, max_x]).range([0, width-right_margin]);
    yScale = d3.scaleLinear().domain([min_y, max_y]).range([height, 0]);
    radiusScale = d3.scaleSqrt().domain([0, d3.max(data, elem => elem["critics"]) || 1]).range([5, 15]);

    elegibleData = data.sort((a, b) => {
        return (b["critics"] + b["users"]) - (a["critics"] + a["users"])
    }).slice(0, maxBubbleCount);

    generateFilters(data);
    enableSearch(data);
    createAxis();
    createBubbles(elegibleData);
});

function handleMouseOver(event, d) {
    transform = d3.zoomTransform(svg.node()); // Obtiene la transformación actual (zoom y traslación)

    d3.select(this)
        .transition()
        .duration(radiusTransition)
        .ease(d3.easeQuadInOut)
        .attr("r", radiusScale(d.critics) * 1.4)

    // Calcula las posiciones del círculo transformadas
    const cx = transform.applyX(xScale(d["user score"]));
    const cy = transform.applyY(yScale(d["score"]));

    // Define la distancia visual fija para el tooltip
    const visualOffsetX = 70; // Distancia horizontal fija en píxeles
    const visualOffsetY = 55; // Distancia vertical fija en píxeles

    // Crea el grupo del tooltip
    const rects = svg.append("g").attr("class", "rects");
    const texts = svg.append("g").attr("class", "texts");

    // Añade el texto al tooltip
    const text = texts.append("text")
        .attr("x", cx + visualOffsetX + 10) // Posiciona el texto con offset
        .attr("y", cy + visualOffsetY + 20)
        .style("font-size", "15px")
        .style("font-family", "Georgia, serif")
        .style("fill", "black")

    text.append("tspan")
        .text(`${d.name}`)
        .style("font-size", "18px")
        .style("font-weight", "bold")
        .attr("x", cx + visualOffsetX + 10)
        .attr("dy", "1.2em")
        .style("opacity", 0)
        .transition()
        .duration(rectAparitionTime)
        .ease(d3.easeQuadOut)
        .style("opacity", 1)

    text.append("tspan")
        .text(`Géneros: ${d.genre.join(", ")}`)
        .attr("x", cx + visualOffsetX + 10)
        .attr("dy", "1.2em")
        .style("opacity", 0)
        .transition()
        .duration(rectAparitionTime)
        .ease(d3.easeQuadOut)
        .style("opacity", 1)

    text.append("tspan")
        .text(`Fecha: ${d["r-date"]}`)
        .attr("x", cx + visualOffsetX + 10)
        .attr("dy", "1.2em")
        .style("opacity", 0)
        .transition()
        .duration(rectAparitionTime)
        .ease(d3.easeQuadOut)
        .style("opacity", 1)

    text.append("tspan")
        .text(`Desarrolladores: ${d.developer}`)
        .attr("x", cx + visualOffsetX + 10)
        .attr("dy", "1.2em")
        .style("opacity", 0)
        .transition()
        .duration(rectAparitionTime)
        .ease(d3.easeQuadOut)
        .style("opacity", 1)

    text.append("tspan")
        .text(`Consola: ${d.platform}`)
        .attr("x", cx + visualOffsetX + 10)
        .attr("dy", "1.2em")
        .style("opacity", 0)
        .transition()
        .duration(rectAparitionTime)
        .ease(d3.easeQuadOut)
        .style("opacity", 1)

    text.append("tspan")
        .text(`Jugadores: ${d.players}`)
        .attr("x", cx + visualOffsetX + 10)
        .attr("dy", "1.2em")
        .style("opacity", 0)
        .transition()
        .duration(rectAparitionTime)
        .ease(d3.easeQuadOut)
        .style("opacity", 1)

    // Usa el bounding box del texto para calcular el tamaño del rectángulo
    const bbox = text.node().getBBox();

    rects.insert("rect")
        .attr("x", bbox.x - 10) // Deja un margen alrededor del texto
        .attr("y", bbox.y - 5)
        .attr("width", bbox.width + 20)
        .attr("height", bbox.height + 10)
        .attr("fill", () => {
            if (colores[selection][d[selection]]){
                return colores[selection][d[selection]];
            } else return "#aaaaaa";
        })
        .attr("rx", 5)
        .attr("ry", 5)
        .style("opacity", 0)
        .transition()
        .duration(rectAparitionTime)
        .ease(d3.easeQuadOut)
        .style("opacity", 0.9)

}

function handleMouseOut(event, d) {
    svg.selectAll(".rects")
        .transition()
        .duration(rectAparitionTime)
        .ease(d3.easeQuadOut)
        .style("opacity", 0)
        .remove();
    svg.selectAll(".texts").remove();

    d3.select(this)
        .transition()
        .duration(radiusTransition/2)
        .ease(d3.easeQuadInOut)
        .attr("r", radiusScale(d.critics))

}

function zoomed(event) {
    transform = event.transform;
    if (transform.k < 1) {
        transform.k = 1.1
        return;
    }

    const newXScale = transform.rescaleX(xScale);
    const newYScale = transform.rescaleY(yScale);

    newXScale.domain([Math.max(min_x, newXScale.domain()[0]), Math.min(max_x, newXScale.domain()[1])]);
    newYScale.domain([Math.max(min_y, newYScale.domain()[0]), Math.min(max_y, newYScale.domain()[1])]);



    translate_group.select(".xAxis").call(d3.axisBottom(newXScale));
    translate_group.select(".yAxis").call(d3.axisLeft(newYScale));

    contentGroup.selectAll('.vertical-grid')
        .attr('x1', d => newXScale(d) + margin_x)
        .attr('x2', d => newXScale(d) + margin_x)

    contentGroup.selectAll('.horizontal-grid')
        .attr('y1', d => newYScale(d) + margin_y)
        .attr('y2', d => newYScale(d) + margin_y)

    scatter.select('.bubbles').selectAll('circle')
        .attr('cx', d => newXScale(d["user score"]) + margin_x)
        .attr('cy', d => newYScale(d["score"]) + margin_y)


}

function createBubbles(data) {
    const bubbleGroup = scatter.append('g')
        .attr('class', 'bubbles');

    const zoom = d3.zoom()
        .scaleExtent([1, 10])
        .translateExtent([[0, 0], [width + 2*margin_x, height + 2*margin_y]])
        .on("zoom", zoomed)
        .filter((event) => event.type === "mousedown" || event.type === "mousemove" || event.type === "wheel");

    bubbleGroup.selectAll("circle")
        .data(data)
        .enter()
        .append("circle")
        .attr("cx", d => xScale(d["user score"]) + margin_x)
        .attr("cy", d => yScale(d["score"]) + margin_y)
        .attr("r",  d => radiusScale(d["critics"]))
        .attr("fill", d => {
            if (colores[selection][d[selection]]){
                return colores[selection][d[selection]];
            } else return "#aaaaaa";
        })
        .attr("stroke", "#000000")
        .attr("stroke-width", 1)
        .lower()
        .on("mouseover", handleMouseOver)
        .on("mouseout", handleMouseOut)

    svg.call(zoom);

}

function removeBubbles() {
    d3.select(".bubbles").remove();
}

function changeSelection (event) {
    removeBubbles();
    selection = event.target.value;
    createBubbles(elegibleData);
}

function generateFilters(data) {
    const filtersDiv = d3.select("#filters");
    filtersDiv.html(""); // Limpia los filtros existentes

    // Obtiene valores únicos para el filtro seleccionado
    // Genera un checkbox para cada valor único
    const generos = [
        "Fantasy",
        "Sports",
        "Open-World",
        "Action",
        "Shooter",
        "Arcade",
        "2D",
        "3D",
        "Rhythm",
        "Team",
        "Roguelike"
    ];

    generos.forEach(value => {
        const filter = filtersDiv.append("div").attr("class", "filter");

        filter.append("input")
            .attr("type", "checkbox")
            .attr("id", `filter-${value}`)
            .attr("value", value)
            .on("change", () => applyFilters(data)); // Aplica los filtros al cambiar

        filter.append("label")
            .attr("for", `filter-${value}`)
            .text(value);
    });
}

function applyFilters(data) {
    let selectedFilters = d3.selectAll("#filters input[type='checkbox']")
        .nodes()
        .filter(input => input.checked)
        .map(input => input.value);


    // Filtra los datos en base a los filtros seleccionados
    elegibleData = data.filter(d => {
        if (selectedFilters.length === 0) return true;
        for (filter of selectedFilters) {
            if (!d.genre.includes(filter)) return false;
        }
        return true;
    }).splice(0, maxBubbleCount);
    // Actualiza la visualización con los datos filtrados
    removeBubbles();
    createBubbles(elegibleData);
}

function enableSearch(data){
    d3.select("#search")
        .on("input", search)

    function search(event) {
        const keyName = event.target.value.toLowerCase();
        removeBubbles();
        elegibleData = data.filter(d => d.name.toLowerCase().includes(keyName)).splice(0, maxBubbleCount);

        createBubbles(elegibleData);
    }

}

